from typing_ex.builtin_typing import (
    Any,
    NamedTuple,
    Mapping,
    MappingProxyType,
    Optional,
    Type,
    TypeAlias,
)
from typing_ex.type_info import TypeInfo

_SCHEMA_ATTR = "_schema"

PropertyDVTTuple = NamedTuple(
    "PropertyDVTTuple", [("default", Any), ("type", TypeInfo)]
)

Schema: TypeAlias = "MappingProxyType[str, PropertyDVTTuple]"


class PropertyValueError(Exception):
    def __init__(self, property_name: str, value: Any, expected_type: TypeInfo):
        self.property_name = property_name
        self.value = value
        self.expected_type = expected_type
        super().__init__(
            f"value for property '{property_name}'"
            f"is not of type {expected_type.name}"
        )


class UnknownPropertyError(Exception):
    def __init__(self, property_name: str):
        self.property_name = property_name
        super().__init__(f"unknown property '{property_name}'")


class ReservedPropertyError(Exception):
    def __init__(self, property_name: str):
        self.property_name = property_name
        super().__init__(f"property '{property_name}' is reserved")


def _ns_handler(ns: dict):
    annotations = {
        k: v for k, v in ns.get("__annotations__", {}).items() if k[0] != "_"
    }  # filter out private members
    if any(isinstance(annotation, str) for annotation in annotations.values()):
        raise RuntimeError(
            "TypedDefaultDict cannot be used with "
            "'from __future__ import annotations'"
        )
    schema = {
        k: PropertyDVTTuple(ns.get(k), TypeInfo[annotations[k]])
        for k in annotations.keys()
    }
    ns = {k: v for k, v in ns.items() if k not in schema}
    ns["_schema"] = MappingProxyType(schema)
    return ns


class TypedDefaultDictMeta(type):
    def __new__(mcs, name, bases, ns):
        if not bases:
            return type.__new__(mcs, name, bases, ns)
        return type.__new__(mcs, name, bases, _ns_handler(ns))

    def __getattr__(mcs, k: str):
        if k not in mcs._schema:
            raise AttributeError(k)
        return mcs._schema[k].default


# no more metaclass conflict warning from mypy
_TypedDefaultDict: Type = type.__new__(type, "TypedDefaultDict", (dict,), {})


class TypedDefaultDict(_TypedDefaultDict, metaclass=TypedDefaultDictMeta):

    @property
    def schema(self) -> Schema:
        return self._schema

    def update(self, fields: Optional[Mapping[str, Any]] = None, /, **kwargs):
        super().update(self._get_fields(fields, **kwargs))

    def on_get_unknown_property(self, k) -> Any:
        raise UnknownPropertyError(k)

    def on_set_unknown_property(self, k: str, v: Any) -> Any:
        raise UnknownPropertyError(k)

    def set(self, k: str, v: Any) -> None:
        super().__setitem__(k, v)

    def __init__(self, fields: Optional[Mapping[str, Any]] = None, /, **kwargs):
        super().__init__(self._get_fields(fields, **kwargs))

    def __getitem__(self, k: str):
        if k in self:
            return super().__getitem__(k)
        if k in self._schema:
            return self._schema[k].default
        return self.on_get_unknown_property(k)

    def __setitem__(self, k: str, v: Any):
        if self._check_property(k):
            self._check_value_type(k, v)
            super().__setitem__(k, v)
            return
        v = self.on_set_unknown_property(k, v)
        if v is not None:
            super().__setitem__(k, v)

    __getattr__ = __getitem__
    __setattr__ = __setitem__

    def _check_property(self, k: str) -> bool:
        if not isinstance(k, str):
            raise TypeError("property name must be a string")
        if k == _SCHEMA_ATTR:
            raise ReservedPropertyError(k)
        return k in self._schema

    def _check_value_type(self, k: str, v: Any):
        if not self._schema[k].type.check_value(v):
            raise PropertyValueError(k, v, self._schema[k].type)

    def _verify_data(self, data: Mapping[str, Any]) -> Mapping[str, Any]:
        remove_keys = []
        for k, v in data.items():
            if self._check_property(k):
                self._check_value_type(k, v)
                continue
            v = self.on_set_unknown_property(k, v)
            if v is None:
                remove_keys.append(k)
            else:
                data[k] = v
        return {k: v for k, v in data.items() if k not in remove_keys}

    def _get_fields(
        self, data: Optional[Mapping[str, Any]] = None, /, **kwargs
    ) -> Mapping[str, Any]:
        if data and kwargs:
            raise ValueError("data and kwargs cannot be used together")
        if data and not isinstance(data, Mapping):
            raise TypeError("data must be a mapping")
        else:
            return self._verify_data(data or kwargs or {})



"""import builtin types"""

import sys
from types import MappingProxyType, DynamicClassAttribute, new_class
from typing import (
    Any,
    NamedTuple,
    NewType,
    Type,
    TypeVar,
    ClassVar,
    Union,
    Mapping,
    Sequence,
    Iterable,
    Callable,
    Generic,
    Optional,
    List,
    Dict,
    Tuple,
    overload,
    get_args,
    get_origin,
    final,
)
from abc import abstractmethod
from typing_extensions import Self, TypeAlias, override

if sys.version_info < (3, 10):
    UnionType = NewType("UnionType", Union)  # dummy type
else:
    from types import UnionType


AnyUnion: TypeAlias = Union[Type, UnionType, object]
AnyType: TypeAlias = Union[Type, AnyUnion, None]
__all__ = [
    "Any",
    "ClassVar",
    "DynamicClassAttribute",
    "Dict",
    "List",
    "NamedTuple",
    "NewType",
    "Tuple",
    "Type",
    "TypeVar",
    "TypeAlias",
    "Union",
    "UnionType",
    "Mapping",
    "Sequence",
    "Iterable",
    "Callable",
    "Generic",
    "Optional",
    "Self",
    "overload",
    "override",
    "get_args",
    "get_origin",
    "new_class",
    "final",
    "abstractmethod",
    "MappingProxyType",
]



from typing_ex.enum_ex import EnumEx


class TestEnum(EnumEx):
    A = 0
    B = 1
    A_ALIAS = A
    A_ALIAS2 = 0
    B_ALIAS = 1
    B_ALIAS2 = B
    C = 456


def test_int_enum():
    assert TestEnum.value_type is int
    assert TestEnum.A == 0
    assert TestEnum.B == 1
    assert TestEnum.C == 456
    assert TestEnum.A_ALIAS == 0
    assert TestEnum.A_ALIAS == TestEnum.A
    assert TestEnum.A_ALIAS2 == TestEnum.A_ALIAS
    assert TestEnum.A.name == "A"
    assert TestEnum.B.name == "B"
    assert TestEnum.A_ALIAS.name == "A_ALIAS"
    assert TestEnum.A_ALIAS2.name == "A_ALIAS2"
    assert TestEnum.A_ALIAS.orig_name == "A"
    assert TestEnum.B_ALIAS.orig_name == "B"
    assert TestEnum.A_ALIAS.origin == TestEnum.A
    assert TestEnum.B_ALIAS.origin == TestEnum.B
    assert TestEnum.A < TestEnum.B
    assert TestEnum.A < TestEnum.C
    assert TestEnum.A != TestEnum.B
    assert TestEnum.A != TestEnum.C
    assert TestEnum.B != TestEnum.C
    assert TestEnum.A == TestEnum.A
    assert TestEnum.C == TestEnum.C
    assert TestEnum.values == (0, 1, 456), TestEnum.values


class TestStrEnum(EnumEx):
    __value_type__ = str
    FOO = "foo"
    BAR = "bar"
    BAZ = "baz"
    FOO_ALIAS = FOO
    BAR_ALIAS = BAR
    BAZ_ALIAS = BAZ

    _private_attr = "abc"

    @property
    def custom_prop(self):
        return 123


def test_value_type():
    assert TestStrEnum.value_type is str


def test_non_enum_attrs():
    assert TestStrEnum._private_attr == "abc"
    assert TestStrEnum.FOO.custom_prop == 123


def test_iter_enum():
    assert tuple(TestStrEnum) == (
        TestStrEnum.FOO,
        TestStrEnum.BAR,
        TestStrEnum.BAZ,
    ), tuple(
        TestStrEnum
    )  # aliases are skipped


def test_names():
    assert TestStrEnum.names == (
        "FOO",
        "BAR",
        "BAZ",
        "FOO_ALIAS",
        "BAR_ALIAS",
        "BAZ_ALIAS",
    ), TestStrEnum.names


def test_values():
    assert TestStrEnum.values == ("foo", "bar", "baz"), TestStrEnum.values


def test_enums():
    assert TestStrEnum.enums == (
        TestStrEnum.FOO,
        TestStrEnum.BAR,
        TestStrEnum.BAZ,
        TestStrEnum.FOO_ALIAS,
        TestStrEnum.BAR_ALIAS,
        TestStrEnum.BAZ_ALIAS,
    ), TestStrEnum.enums



from __future__ import annotations
from typing_ex.builtin_typing import Generic, Mapping, Sequence, TypeVar, Dict

_T = TypeVar("_T")
_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


def _immutable(self, *_, **__):
    raise RuntimeError("invalid assignment on frozen type")


class FrozenType(Generic[_T]):
    pass


class FrozenList(list, FrozenType[list]):
    __setitem__ = _immutable
    __delitem__ = _immutable
    pop = _immutable
    remove = _immutable
    append = _immutable
    clear = _immutable
    extend = _immutable
    insert = _immutable
    reverse = _immutable
    sort = _immutable


class FrozenDict(Dict[_KT, _VT], FrozenType[Dict[_KT, _VT]], Generic[_KT, _VT]):
    __setitem__ = _immutable
    __delitem__ = _immutable
    pop = _immutable
    popitem = _immutable
    clear = _immutable
    update = _immutable
    setdefault = _immutable


def frozen_copy(obj):
    if isinstance(obj, Mapping):
        return FrozenDict(obj)
    elif isinstance(obj, list):
        return FrozenList(obj)
    elif isinstance(obj, set):
        return frozenset(obj)
    else:
        return obj



from typing_ex.frozen import FrozenDict, FrozenList, frozen_copy
import pytest


def test_frozen_list():
    assert frozen_copy([1, 2, 3]) == [1, 2, 3]
    assert frozen_copy((1, 2, 3)) == (1, 2, 3)
    assert frozen_copy({1, 2, 3}) == {1, 2, 3}
    assert frozen_copy({1: 1, 2: 2, 3: 3}) == {1: 1, 2: 2, 3: 3}

    frozen_list = FrozenList([1, 2, 3])
    assert frozen_copy(frozen_list) == frozen_list
    assert frozen_copy(frozen_copy(frozen_list)) == frozen_list
    with pytest.raises(RuntimeError):
        frozen_list.append(1)  # appending to frozen list is forbidden
    with pytest.raises(RuntimeError):
        frozen_list.extend([1, 2])  # extending frozen list is forbidden
    with pytest.raises(RuntimeError):
        frozen_list.insert(0, 1)  # inserting to frozen list is forbidden
    with pytest.raises(RuntimeError):
        frozen_list.pop()  # popping from frozen list is forbidden
    with pytest.raises(RuntimeError):
        frozen_list.remove(1)  # removing from frozen list is forbidden
    with pytest.raises(RuntimeError):
        frozen_list.reverse()  # reversing frozen list is forbidden
    with pytest.raises(RuntimeError):
        frozen_list.sort()  # sorting frozen list is forbidden


def test_frozen_dict():
    frozen_dict = FrozenDict({"a": 1, "b": 2})
    assert frozen_copy(frozen_dict) == frozen_dict
    assert frozen_copy(frozen_copy(frozen_dict)) == frozen_dict
    with pytest.raises(RuntimeError):
        frozen_dict["a"] = 1  # appending to frozen dict is forbidden
    with pytest.raises(RuntimeError):
        frozen_dict.update({"a": 1})  # updating frozen dict is forbidden
    with pytest.raises(RuntimeError):
        frozen_dict.pop("a")  # popping from frozen dict is forbidden
    with pytest.raises(RuntimeError):
        frozen_dict.popitem()  # popping from frozen dict is forbidden
    with pytest.raises(RuntimeError):
        frozen_dict.clear()  # clearing frozen dict is forbiddenn
    with pytest.raises(RuntimeError):
        frozen_dict.setdefault("a")



"""Extended typing module"""
from typing_ex.type_info import TypeInfo
from typing_ex.enum_ex import EnumEx
from typing_ex.typed_defaultdict import (
    TypedDefaultDict,
    PropertyDVTTuple,
    PropertyValueError,
    UnknownPropertyError,
    ReservedPropertyError,
)
from typing_ex.frozen import FrozenType, FrozenList, FrozenDict

__all__ = [
    "TypeInfo",
    "EnumEx",
    "TypedDefaultDict",
    "PropertyDVTTuple",
    "PropertyValueError",
    "UnknownPropertyError",
    "ReservedPropertyError",
    "FrozenType",
    "FrozenList",
    "FrozenDict",
]



from __future__ import annotations
from typing_ex.type_info import TypeInfo, Union
from typing_ex.builtin_typing import UnionType
import sys


def test_check_union():
    if sys.version_info >= (3, 10):
        assert TypeInfo.check_union(int | str, str | int)
        assert TypeInfo.check_union(int, int | str)
        assert not TypeInfo.check_union(int | str, str)
        assert not TypeInfo.check_union(int | str, int | float)
        assert TypeInfo.check_union(Union[int, str], str | int)
        assert TypeInfo.check_union(int, Union[int | str])
        assert not TypeInfo.check_union(Union[int, str], int | float)

    assert TypeInfo.check_union(Union[int, str], Union[str, int])
    assert not TypeInfo.check_union(Union[int, str], str)
    assert not TypeInfo.check_union(Union[int, str], Union[int, float])
    assert not TypeInfo.check_union(Union[int, str], Union[int, float])


def test_check_value():
    if sys.version_info >= (3, 10):
        assert TypeInfo[dict[int, str]].check_value({1: "2"})
        assert TypeInfo[dict[str, int | str]].check_value({"1": 1, "2": "2"})
        assert TypeInfo[list[int | str]].check_value([1, "2", "3"])
        assert not TypeInfo[int | str].check_value(1.0)
        assert not TypeInfo[dict[int, str]].check_value({"2": "2"})
        assert not TypeInfo[dict[int, int | str]].check_value({"2": 2, "3": "3"})
        assert not TypeInfo[dict[int, str]].check_value({1.0: "2"})
        assert not (TypeInfo[list[int | str]].check_value([1.0, "2", "3"]))
    assert TypeInfo[int].check_value(1)
    assert TypeInfo[Union[int, str]].check_value(1)
    assert TypeInfo[Union[int, str]].check_value("1")
    assert not TypeInfo[int].check_value(1.0)


def test_type_name():
    if sys.version_info >= (3, 10):
        assert TypeInfo[int | str].name == "int | str"
        assert TypeInfo[dict[int, float]].name == "dict[int, float]"
        assert TypeInfo[Union[int, str]].name == "int | str"
    else:
        assert TypeInfo[Union[int, str]].name == "typing.Union[int, str]"
    assert TypeInfo[int].name == "int"
    assert TypeInfo[None].name == "None"
    assert str(TypeInfo[str]) == "str"


def test_type_origin():
    if sys.version_info >= (3, 10):
        assert TypeInfo[int | str].origin is UnionType
        assert TypeInfo[dict[int, str]].origin is dict
        assert TypeInfo[Union[int, str]].origin is UnionType
    else:
        assert TypeInfo[Union[int, str]].origin is Union
    assert TypeInfo[int].origin is int
    assert TypeInfo[None].origin is None


def test_type_args():
    if sys.version_info >= (3, 10):
        assert TypeInfo[int | str].args == (int, str)
        assert TypeInfo[dict[int, float]].args == (int, float)
    assert TypeInfo[int].args == ()
    assert TypeInfo[Union[int, float]].args == (int, float)
    assert TypeInfo[None].args == ()



import sys
from typing_ex.builtin_typing import (
    Type,
    TypeVar,
    UnionType,
    Union,
    MappingProxyType,
    Iterable,
    Callable,
    Mapping,
    Sequence,
    Tuple,
    Dict,
    Any,
    AnyType,
    AnyUnion,
    get_args,
    get_origin,
)

_ValueType = TypeVar("_ValueType")


class TypeInfoMeta(type):
    @staticmethod
    def __getitem__(t):
        return TypeInfo.get(t)


class TypeInfo(type, metaclass=TypeInfoMeta):
    _instances: Dict[type, "TypeInfo"] = {}
    _type_name_dict: Dict[type, str] = {}
    _info: Dict[str, Any]

    __test__ = False

    @classmethod
    def get(mcs, t):
        if t in TypeInfo._instances:
            return TypeInfo._instances[t]
        origin = get_origin(t)
        args = get_args(t)
        info = MappingProxyType(
            {
                "type": t,
                "origin": origin or t,
                "args": args,
            }
        )
        obj = type.__new__(
            mcs,
            TypeInfo._type_name(t, origin, args),
            (type,),
            {"_info": info},
        )
        mcs._instances[t] = obj
        return obj

    @property
    def type(cls) -> Type:
        return cls._info["type"]

    @property
    def origin(cls) -> Type:
        return cls._info["origin"]

    @property
    def args(cls) -> Tuple[Any, ...]:
        return cls._info["args"]

    @property
    def name(cls) -> str:
        return cls._type_name(cls.type, cls.origin, cls.args)

    @property
    def is_union_type(cls) -> bool:
        return cls._is_union_type(cls.type, cls.origin)

    @property
    def is_mapping(cls) -> bool:
        return issubclass(cls.origin, Mapping)

    @property
    def is_set(cls) -> bool:
        return issubclass(cls.origin, set)

    @property
    def is_sequence(cls) -> bool:
        return issubclass(cls.origin, Sequence)

    @staticmethod
    def check_union(t1: AnyUnion, t2: AnyUnion) -> bool:
        if TypeInfo._is_union_type(t1, t1):
            t1_args = get_args(t1)
            t2_args = get_args(t2)
            return len(t1_args) == len(t2_args) and all(
                arg in t2_args for arg in t1_args
            )
        if TypeInfo._is_union_type(t2, t2):
            return t1 in get_args(t2)
        return False

    def check_value(cls, value) -> bool:
        if value is None:
            return cls.type is None
        if cls.is_union_type:
            return issubclass(type(value), cls.args)
        try:
            if not isinstance(value, cls.type):
                return False
        except TypeError:
            if not isinstance(value, cls.origin):
                return False
        if not isinstance(value, Iterable):
            return not cls.args
        for item in cls._iter_value(value):
            if not cls._check_values(cls.args, item):
                return False
        return True

    def __iter__(cls):
        for k, v in cls._info.items():
            yield f"{k} = {v}"

    def __str__(cls) -> str:
        return str(cls.name)

    __repr__ = __str__

    @staticmethod
    def _check_args(
        args1: Tuple[AnyType, ...],
        types_or_values: Tuple[_ValueType, ...],
        check_fn: Callable[[AnyType, _ValueType], bool],
    ):
        if not args1:
            return True
        if not types_or_values:
            return False
        return all(
            check_fn(arg_a, arg_b) for arg_a, arg_b in zip(args1, types_or_values)
        )

    @staticmethod
    def _check_values(args: Tuple[AnyType, ...], values: Tuple[_ValueType, ...]):
        return TypeInfo._check_args(
            args, values, lambda arg, value: TypeInfo[arg].check_value(value)
        )

    def _iter_value(cls, value: Any):
        assert isinstance(value, Iterable)
        if isinstance(value, Mapping):
            it = iter(value.items())
        else:
            it = iter(value)
        for item in it:
            yield item if isinstance(item, Iterable) else [item]

    @staticmethod
    def _type_name(t, origin: AnyType, args: Tuple[AnyType, ...]):
        if t is None:
            return "None"
        if t in TypeInfo._type_name_dict:
            return TypeInfo._type_name_dict[t]
        if not origin and not args:
            t_name = t.__name__
        else:
            t_name = str(t)
        TypeInfo._type_name_dict[t] = t_name
        return t_name

    @staticmethod
    def _is_union_type(t_cls: AnyType, t_origin: AnyType):
        if sys.version_info >= (3, 10):
            return isinstance(t_cls, UnionType) or get_origin(t_origin) is Union
        return get_origin(t_cls) is Union or get_origin(t_origin) is Union



from enum import Enum
from typing_ex.builtin_typing import (
    DynamicClassAttribute,
    Type,
    ClassVar,
    Dict,
    Tuple,
    final,
    Any,
    abstractmethod,
    new_class,
)


"""
Trick static type checker
to recognize enum as EnumEx instead of __value_type__
"""


class EnumEx(Enum):
    names: ClassVar[Tuple[str]]
    values: ClassVar[Tuple]
    enums: ClassVar[Tuple["_EnumEx"]]
    value_type: ClassVar[Type]

    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    @abstractmethod
    def value(self): ...

    @property
    @abstractmethod
    def orig_name(self) -> str: ...

    @property
    @abstractmethod
    def is_alias(self) -> bool: ...


class _ClassStorage:
    name_value_dict: Dict[str, Any]
    value_name_dict: Dict[Any, str]
    name_orig_name_dict: Dict[str, str]
    value_type: Type
    instances: Dict[str, "EnumEx"]
    class_: Type

    def __init__(self, value_type: Type, nv_dict: Dict[str, Any]) -> None:
        self.name_value_dict = nv_dict
        self.value_name_dict = {}
        self.name_orig_name_dict = {}
        self.instances = {}
        self.value_type = value_type
        for k, v in nv_dict.items():
            if not isinstance(v, value_type):
                raise ValueError(
                    f"Enum {k} has a value {v} "
                    "which is not of type of {value_type.__name__}"
                )
            if v not in self.value_name_dict:
                self.value_name_dict[v] = k
                self.name_orig_name_dict[k] = k
                continue
            self.name_orig_name_dict[k] = self.value_name_dict[v]

    def init_instances(self):
        for k, v in self.name_value_dict.items():
            self.instances[k] = self.class_.__create__(
                k, v, self.name_orig_name_dict[k]
            )


class EnumExMeta(type):
    def __new__(mcs, name: str, bases: Tuple[type, ...], attrs: Dict[str, Any]):
        if name == "EnumExMeta":
            return type.__new__(mcs, name, bases, attrs)
        value_type = attrs.pop("__value_type__", None) or int
        nv_dict = {}
        for attr, v in attrs.items():
            if attr[0] != "_" and attr.isupper():
                nv_dict[attr] = v
        attrs = {k: v for k, v in attrs.items() if k not in nv_dict}
        attrs["__storage__"] = _ClassStorage(value_type, nv_dict)
        tp_enum = type.__new__(mcs, name, bases, attrs)
        tp_enum.__storage__.class_ = tp_enum
        tp_enum.__storage__.init_instances()
        return tp_enum

    def __getattr__(mcs, k: str):
        cls_storage = mcs.__storage__
        if k not in cls_storage.instances:
            raise AttributeError(k)
        return cls_storage.instances[k]

    def __iter__(mcs):
        """Iterate through non-alias enums"""
        cls_storage = mcs.__storage__
        for v in cls_storage.instances.values():
            if v.name == v.orig_name:
                yield v

    @DynamicClassAttribute
    def names(cls):
        return tuple(cls.__storage__.instances.keys())

    @DynamicClassAttribute
    def values(cls):
        return tuple(cls.__storage__.value_name_dict.keys())

    @DynamicClassAttribute
    def enums(cls):
        return tuple(cls.__storage__.instances.values())

    @DynamicClassAttribute
    def value_type(cls) -> Type:
        """all enums"""
        return cls.__storage__.value_type


class _EnumEx(metaclass=EnumExMeta):
    __slots__ = ("_name", "_value", "_orig_name")
    __value_type__: ClassVar[Type]
    __test__: ClassVar[bool] = False

    names: ClassVar[Tuple[str]]
    values: ClassVar[Tuple]
    enums: ClassVar[Tuple["EnumEx"]]
    value_type: ClassVar[Type]

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self):
        return self._value

    @property
    def orig_name(self) -> str:
        return self._orig_name

    @property
    def origin(self) -> "EnumEx":
        return self.__class__.__storage__.instances[self._orig_name]

    @property
    def is_alias(self) -> bool:
        return self._orig_name != self._name

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self._value == other._value
        if isinstance(other, self.__class__.value_type):
            return self._value == other
        raise ValueError(
            f"cannot compare enum of {self.__class__.value_type.__name__} "
            f"with {other.__class__.__name__}"
        )

    def __lt__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self._value < other._value
        if isinstance(other, self.__class__.value_type):
            return self._value < other
        raise ValueError(
            f"cannot compare enum of {self.__class__.value_type.__name__} "
            f"with {other.__class__.__name__}"
        )

    def __hash__(self) -> int:
        return hash(f"{self.name}:{self.value}")

    def __str__(self) -> str:
        return self.name

    __repr__ = __str__

    @final
    @classmethod
    def __create__(cls, *args):
        obj = object.__new__(cls)
        cls.__init__(obj, *args)
        return obj

    @final
    def __init__(self, name: str, value, orig_name: str) -> None:
        self._name = name
        self._value = value
        self._orig_name = orig_name


EnumEx = new_class("EnumEx", (_EnumEx,))  # type: ignore[misc,assignment] # noqa: F811
EnumEx.__doc__ = Enum.__doc__



import pytest
from typing_ex.typed_defaultdict import PropertyValueError, TypedDefaultDict
from typing_ex.builtin_typing import Any, Optional, Dict, List, override


class DictDefault(TypedDefaultDict):
    image: Optional[str] = None
    mounts: List[str] = []
    foo: str = "bar"


class DictNoDV(TypedDefaultDict):
    foo: str
    bar: int


def test_default_values():
    assert DictDefault.image is None
    assert DictDefault.mounts == []
    assert DictDefault.foo == "bar"
    assert DictNoDV.foo is None
    assert DictNoDV.bar is None


def test_init_getattr_getitem():
    d = DictDefault(image="foo", mounts=["foo", "bar"])
    assert d["image"] == "foo", d["image"]
    assert d["foo"] == "bar", d["foo"]  # default value
    assert d["mounts"] == ["foo", "bar"], d["mounts"]
    assert d.image == "foo", d.image
    assert d.foo == "bar", d.foo  # default value
    assert d.mounts == ["foo", "bar"], d.mounts


def test_invalid_property_type():
    with pytest.raises(PropertyValueError):
        DictDefault(image="foo", mounts=["foo", 1])


class DictWithPropertiesAndMethods(TypedDefaultDict):
    foo: str = "foo"
    bar: str = "bar"

    _extra_props: Dict[str, Any] = {}

    @property
    def foo_bar(self) -> str:
        return self.foo + self.bar

    def foo_bar_method(self) -> str:
        return self.foo + self.bar

    @classmethod
    def foo_bar_cls_method(cls) -> str:
        return cls.foo + cls.bar

    @staticmethod
    def foo_bar_static_method() -> str:
        return DictWithPropertiesAndMethods.foo + DictWithPropertiesAndMethods.bar

    @override
    def on_get_unknown_property(self, k) -> Any:
        raise RuntimeError(f"hello {k}")

    @override
    def on_set_unknown_property(self, k: str, v: Any):
        self._extra_props[k] = v


def test_properties_and_methods():

    assert DictWithPropertiesAndMethods.foo_bar_cls_method() == "foobar"
    assert DictWithPropertiesAndMethods.foo_bar_static_method() == "foobar"
    test_dict = DictWithPropertiesAndMethods(foo="bar", bar="foo")
    assert test_dict.foo_bar == "barfoo"
    assert test_dict.foo_bar_method() == "barfoo"
    assert test_dict.foo_bar_cls_method() == "foobar"
    assert test_dict.foo_bar_static_method() == "foobar"


def test_overridden_set_unknown_prop():
    test_dict = DictWithPropertiesAndMethods(foo="bar", bar="foo")
    test_dict.update(baz="baz", baf="baf")
    assert test_dict._extra_props["baz"] == "baz"
    assert test_dict._extra_props["baf"] == "baf"
    test_dict.update({"bar": "bararara", "bab": 2})
    assert test_dict.bar == "bararara"
    assert test_dict._extra_props["bab"] == 2


def test_overridden_get_unknown_prop():
    test_dict = DictWithPropertiesAndMethods()
    with pytest.raises(RuntimeError, match="hello baz"):
        baz = test_dict.baz
    with pytest.raises(RuntimeError, match="hello baf"):
        baf = test_dict["baf"]


def test_set():
    test_dict = DictWithPropertiesAndMethods()
    test_dict.set("blablabla", {"a": 1, "b": 2})
    assert test_dict.blablabla["a"] == 1
    assert test_dict.blablabla["b"] == 2
